"""
=======
Helpers
=======
Useful helper methods that frequently used in this project
"""

__all__ = ['custom_exception_handler']

import inspect
import logging
import mimetypes
import os
import re
import traceback

import sys
from django.conf import settings
from django.http import HttpResponse
from rest_framework.views import exception_handler
from sbc_drf.errors import ErrorMessage

from . import errors as err

L = logging.getLogger('app.' + __name__)

first_cap_re = re.compile('(.)([A-Z][a-z]+)')
all_cap_re = re.compile('([a-z0-9])([A-Z])')


def custom_exception_handler(exc, context):
    """
    Replace rest_framework's default handler to generate error response as per API specification
    """

    #: Call REST framework's default exception handler first,
    #: to get the standard error response.
    response = exception_handler(exc, context)
    request = context.get('request')

    # Getting originating location of exception
    frm = inspect.trace()[-1]
    mod = inspect.getmodule(frm[0])
    func = traceback.extract_tb(sys.exc_info()[2])[-1][2]
    log_extra = {'request': request, 'request_id': request.id,
                 'culprit': "%s in %s" % (mod.__name__, func)}

    if response is None and settings.DEBUG:
        return response

    if response:
        L.warning(exc, extra=log_extra, exc_info=True)
    else:
        L.exception(exc, extra=log_extra)

    # Response is none, meaning builtin error handler failed to generate response and it needs to
    #  be converted to json response
    if response is None:
        return custom_exception_handler(err.APIException(*ErrorMessage.UNEXPECTED), context)

    # Passing context to help the renderer to identify if response is error or normal data
    if isinstance(response.data, list):
        response.data = response.data[0]

    response.data['_context'] = 'error'
    response.data['type'] = exc.__class__.__name__
    response.data['status_code'] = response.status_code
    response.data['error_code'] = getattr(exc, 'error_code', 0)

    return response


def attachment_response(file, request, remove_file=True):
    filename = os.path.basename(file)
    with open(file, 'rb') as fp:
        response = HttpResponse(fp.read())

    type_, encoding = mimetypes.guess_type(file)
    if type_ is None:
        type_ = 'application/octet-stream'

    response['Content-Type'] = type_
    response['Content-Length'] = str(os.stat(file).st_size)
    if encoding is not None:
        response['Content-Encoding'] = encoding

    # To inspect details for the below code, see http://greenbytes.de/tech/tc2231/
    if u'WebKit' in request.META['HTTP_USER_AGENT']:
        # Safari 3.0 and Chrome 2.0 accepts UTF-8 encoded string directly.
        filename_header = 'filename=%s' % filename
    elif u'MSIE' in request.META['HTTP_USER_AGENT']:
        # IE does not support internationalized filename at all.
        # It can only recognize internationalized URL, so we do the trick via routing rules.
        filename_header = ''
    else:
        # For others like Firefox, we follow RFC2231 (encoding extension in HTTP headers).
        filename_header = 'filename*=UTF-8\'\'%s' % filename

    response['Content-Disposition'] = 'attachment; ' + filename_header
    if remove_file:
        os.unlink(file)

    return response



"""
=======
Routers
=======
"""
import re

from django.core.exceptions import ImproperlyConfigured
from rest_framework import routers as rf_routers
from rest_framework_extensions import routers as rfe_routers


class SimpleRouter(rf_routers.SimpleRouter):
    """
    Extends simple router with default trailing_slash to False.

    It also converts endpoint name starts with ``action?_`` to ``/actions/...``
    """

    def __init__(self):
        super().__init__(trailing_slash=False)

    def get_urls(self):
        urls = super().get_urls()

        # Replace action urls with slashed name
        for url in urls:
            url.pattern._regex = re.sub(r'/actions?_', '/actions/', url.pattern._regex, 1)

        return urls


class ExtendedSimpleRouter(SimpleRouter, rfe_routers.ExtendedSimpleRouter):
    def get_dynamic_routes(self, viewset):
        known_actions = self.get_known_actions()
        dynamic_routes = []
        for methodname in dir(viewset):

            attr = getattr(viewset, methodname)
            httpmethods = getattr(attr, 'bind_to_methods', None)
            if httpmethods:
                endpoint = getattr(attr, 'endpoint', methodname)
                is_for_list = not getattr(attr, 'detail', True)
                if endpoint in known_actions:
                    raise ImproperlyConfigured('Cannot use @action or @link decorator on '
                                               'method "%s" as %s is an existing route'
                                               % (methodname, endpoint))
                httpmethods = [method.lower() for method in httpmethods]
                dynamic_routes.append((httpmethods, methodname, endpoint, is_for_list))
        return dynamic_routes



__all__ = ['ActionSerializerViewSetMixin', 'NestedViewSetMixin']

from collections import deque

from django.http import QueryDict
from rest_framework_extensions.mixins import NestedViewSetMixin as __NestedViewSetMixin


class ActionSerializerViewSetMixin:
    """
    a view-set mixin that returns serializer class based on action being called if specified.

    For eg.: In a view-set if following is defined

        serializer_class = MySerializer
        list_serializer_class = MyListSerializer

        serializer class `MyListSerializer` will be used when list() action
         (usually http GET for list objects.) is called.
    """

    def get_serializer_class(self):
        serializer_class = getattr(self, '%s_serializer_class' % self.action, None)

        if serializer_class:
            return serializer_class
        else:
            return super().get_serializer_class()


class NestedViewSetMixin(__NestedViewSetMixin):
    """
    Extends feature to http://chibisov.github.io/drf-extensions/docs/#nested-router-mixin

    * Method to get parent object
    * Add parent query dict to `request.data`
    """

    def create(self, request, *args, **kwargs):
        """
        Extends to inject parent object id to the request data.
        """
        self.get_parent_object()

        # Since QueryDict now immutable
        if isinstance(request.data, QueryDict):
            request.data._mutable = True

        if type(request.data) in [list, tuple]:
            deque(map(lambda data: data.update(**self.get_parents_query_dict()), request.data))
        else:
            request.data.update(**self.get_parents_query_dict())

        request.parent_query_dict = self.get_parents_query_dict()

        return super().create(request, *args, **kwargs)

    def get_parent_object(self):
        """
        Checks the object's parent object permission and returns it
        """
        # Getting related field name
        parent_object_name = None

        # Spliting direct foreign key and nested foreign keys; later we'll have to query indirect
        # foreign key for permission
        foreign_keys = []
        for key in self.get_parents_query_dict().keys():
            if '__' in key:
                foreign_keys += key.split('__')[1:]
            else:
                parent_object_name = key

        # Getting parent model
        parent_model = self.get_queryset().model._meta.get_field(parent_object_name).related_model
        # Getting parent object
        parent_object = parent_model.objects.get(
            id=self.get_parents_query_dict().get(parent_object_name)
        )

        self.check_object_permissions(self.request, parent_object)

        __parent_object = parent_object
        for key in foreign_keys:
            __parent_object = getattr(__parent_object, key)
            self.check_object_permissions(self.request, __parent_object)

        return parent_object



"""
========
Dispatch
========
Identical to django.dispatch module but adds few more features
"""
import django.dispatch

from celery import shared_task


def async_receiver(signal, sender=None, **kwargs):
    """
    Decorator to perform django signal asynchronously using Celery. The function decorated with
    this should be recognized by celery. django signal mechanism should be working normally and
    no additional changes are required while using in-built signals or custom signals.
    """

    def _decorator(func):
        # Convert normal function to celery task
        func_celery = shared_task(func, **kwargs)

        # Connect to a signal
        if isinstance(signal, (list, tuple)):
            for s in signal:
                # Weak is false as func_celery doesn't exists outside the closure scope. So cannot
                # be referenced weakly and will be erased by garbage collector
                s.connect(func_celery.delay, sender=sender)
        else:
            signal.connect(func_celery.delay, sender=sender)

        # To let celery recognize normal function as celery task
        return func_celery

    return _decorator


def reducer(self):
    return django.dispatch.Signal, (self.providing_args,)


# Monkey patched Signal class to remove non-pickleable attribute
django.dispatch.Signal.__reduce__ = reducer



import logging

from django.conf import settings
from django.contrib.auth import get_user_model
from rest_framework import status
from rest_framework.test import APIClient as APIClient_, APITestCase


class APIClient(APIClient_):
    def generic(self, method, path, data='',
                content_type='application/octet-stream', secure=False, **extra):

        resp = super(APIClient, self).generic(method, path, data=data, content_type=content_type,
                                              secure=secure, **extra)
        return self._fix_response(resp)

    @classmethod
    def _fix_response(cls, response):
        if response.get('Content-Type') != 'application/json':
            return response

        try:
            resp = response.json()
            response.meta = resp.get('meta')
            response.data = resp.get('data') or resp.get('error')
        except:
            pass

        return response


class TestCase(APITestCase):
    client_class = APIClient

    def __init__(self, methodName='runTest'):
        super().__init__(methodName=methodName)

    def get_client(self, user, http_authorization_prefix='Token'):
        c = APIClient()
        c.credentials(HTTP_AUTHORIZATION='%s %s' % (http_authorization_prefix, user.auth_token.key))
        return c

    def mock_timezone_now(self, time):
        try:
            from mock import patch
        except ImportError:
            raise ModuleNotFoundError

        self.patcher = patch('django.utils.timezone.now', lambda: time)
        self.addCleanup(self.patcher.stop)
        self.patcher.start()

    def setUp(self):
        super().setUp()
        settings.DEBUG = True

        self.status_code = status

        logging.disable(logging.NOTSET)


class SetupUserTestCaseMixin:
    admin_client = None
    admin_token = None
    admin = None
    user_client = None
    user_token = None
    user = None

    PWD = 'ucRLRlkwKj8o'

    def setUp(self):
        from rest_framework.authtoken.models import Token

        super().setUp()

        UserModel = get_user_model()

        self.admin = UserModel.objects.create_superuser(email='admin@stickboy.io',
                                                        password=self.PWD)
        self.user = UserModel.objects.create(email='user@stickboy.io', password=self.PWD)

        Token.objects.get_or_create(user=self.admin)
        Token.objects.get_or_create(user=self.user)

        self.admin_client = self.get_client(self.admin)
        self.user_client = self.get_client(self.user)



from django.db.models import Q
from rest_framework import filters


class OwnerOrStaffFilterBackend(filters.BaseFilterBackend):
    """
    Filters the objects based on the list of User foreign key attribute defined in
    `ownership_filter_fields`.

    It checks object's attributes against user id. All staff/admin user's are excluded from being
    filtered.

    As a matter of organization's staff/admin seeing other user's data is prevented with
    `OrganizationFilterBackend`
    """

    def filter_queryset(self, request, queryset, view):
        ownership_filter_fields = getattr(view, 'ownership_filter_fields', None)
        user = request.user

        # ignore when fields are missing
        if not ownership_filter_fields:
            return queryset

        # Allow system's staff/admin
        if user.is_superuser or user.is_staff:
            return queryset

        q = Q()
        for field in ownership_filter_fields:
            q |= Q(**{field: user.id})

        # The reason behind not directly filtering the original queryset with the filter fields
        # it gets duplicated when there's OneToMany fields are being filtered.
        # Distinct can be added but it'll become trickier with custom ordering as 'id' field
        # would always be present in ordering fields
        qs = queryset.model.objects.filter(q).values_list('id', flat=True)
        queryset = queryset.filter(id__in=qs)

        return queryset



"""
======
Errors
======
This module holds exceptions for all other app
"""

from rest_framework import exceptions, status


class ErrorMessage:
    UNEXPECTED = "Unexpected error. We'll fix it soon!", -1


class APIException(exceptions.APIException):
    """
    Exception class that caught by renderer and produce pretty output.

    It also has ``error_code`` attribute that may be set by other app otherwise it'll be ``-1``
    """

    def __init__(self, detail=None, code=None, param=None):
        if isinstance(param, dict):
            detail = detail.format(**param)

        super().__init__(detail=detail, code=code)


class NotFound(APIException):
    status_code = status.HTTP_404_NOT_FOUND


class ValidationError(APIException):
    status_code = status.HTTP_400_BAD_REQUEST



"""
========
Handlers
========
Holds logger handler classes
"""

from raven.contrib.django.raven_compat import handlers


class SentryHandler(handlers.SentryHandler):
    """
    Extends SentryHandler feature that adds thread id to record
    """

    def _emit(self, record):
        request = getattr(record, 'request', None)
        request_id = None
        if request:
            request_id = getattr(request, 'id', None)

        tags = getattr(record, 'tags', {})
        tags.update({
            'thread': record.thread,
            'request_id': request_id
        })
        record.tags = tags

        return super(SentryHandler, self)._emit(record)



"""
Pagination
==========
"""

from rest_framework.pagination import PageNumberPagination as PageNumberPagination_


class PageNumberPagination(PageNumberPagination_):
    """
    Extends standard drf.PageNumberPagination class to enabled page_size param.

    if `view.zero_page_size` is True, it'll allow page_size=0 and pagination for the request
    will be disabled.
    """
    page_size_query_param = 'page_size'
    max_page_size = 100

    def paginate_queryset(self, queryset, request, view=None):
        zero_page_size = getattr(view, 'zero_page_size', False)
        actual_page_size = self.get_actual_page_size(request)

        if zero_page_size is True and actual_page_size == 0:
            self.page_size = None

        return super().paginate_queryset(queryset, request, view)

    def get_actual_page_size(self, request):
        if self.page_size_query_param:
            try:
                return int(request.query_params[self.page_size_query_param])
            except (KeyError, ValueError):
                pass

        return self.page_size


class NoPagination(PageNumberPagination_):
    page_size = None



"""
===========
Permissions
===========
"""
import logging

from rest_framework import permissions as rf_permissions

L = logging.getLogger(__name__)


class DjangoModelPermissions(rf_permissions.DjangoModelPermissions):
    """
    Extends to add view permission
    """
    perms_map = {
        'GET': ['%(app_label)s.view_%(model_name)s'],
        'OPTIONS': [],
        'HEAD': [],
        'POST': ['%(app_label)s.add_%(model_name)s'],
        'PUT': ['%(app_label)s.change_%(model_name)s'],
        'PATCH': ['%(app_label)s.change_%(model_name)s'],
        'DELETE': ['%(app_label)s.delete_%(model_name)s'],
    }


class IsOwnerPermissions(rf_permissions.BasePermission):
    """
    Permission class that grant permission according to object owner.

    :param bool ownership_fields: List of attribute of model property that specify the ownership \
    of object, If it's an empty object, ``True`` will always return.

    .. note:: Condition check on ``ownership_fields`` will be done in ORing fashion
    """

    def has_object_permission(self, request, view, obj):
        """
        Checks for the single object if its user is same as obj. If you want to skip the owner
        permission for LISTing the data, you can set `skip_owner_filter=True` in viewset class
        """
        ownership_fields = getattr(view, 'ownership_fields', None)

        # If it's explicitly mentioned to empty
        if not ownership_fields or request.method == 'GET':
            return True

        _u = request.user

        # Requesting user is accessing to himself?
        if _u == obj:
            return True

        result = any(_u == self._get(obj, field) or _u.id == self._get(obj, field)
                     for field in ownership_fields)
        if result is False:
            logging.warning('Permission denied', extra={'user': _u,
                                                        'ownership_fields': ownership_fields,
                                                        'object': obj})

        return result

    @staticmethod
    def _get(obj, path):
        paths = path.split('__')

        for p in paths:
            obj = getattr(obj, p, None)
            if obj is None:
                return

        return obj


class ActionPermissions(rf_permissions.BasePermission):
    """
    Permission class that checks for custom permissions defined under model's meta option.

    Model permission name should match with the custom action defined in viewset class.
    for eg, if ``action_pay()`` defined under viewset, model permission set should contain custom
    permission name `action_pay`.

    ``has_permission()`` will match for ``<app_label>.<action_name>`` as permission name if user
    has.

    .. note:: Be careful using this class as it doesn't check anything apart from custom \
    permissions.
    """

    def has_permission(self, request, view):
        model_cls = getattr(view, 'model', None)
        queryset = getattr(view, 'queryset', None)

        if model_cls is None and queryset is not None:
            model_cls = queryset.model

        expected_perm = "%(app_label)s.%(action_name)s" % {'app_label': model_cls._meta.app_label,
                                                           'action_name': view.action}

        return request.user.has_perm(expected_perm)



__all__ = ['AddUpdateTimeModelMixin', 'ViewPermModelMetaMixin']

from django.db import models
from django.db.models import Model
from django.utils.translation import ugettext as _


class AddUpdateTimeModelMixin(Model):
    """
    Adds created_at and updated_at fields to a model
    """
    created_at = models.DateTimeField(_('Created at'), auto_now_add=True)
    updated_at = models.DateTimeField(_('Updated at'), auto_now=True)

    class Meta:
        abstract = True


class ViewPermModelMetaMixin:
    """
    Adds view permission to a model along with other default permissions.
    """
    default_permissions = ('add', 'change', 'delete', 'view')



__all__ = ['DynamicFieldsSerializerMixin', 'ExcludeOnUpdateSerializerMixin',
           'InjectReqUserSerializerMixin']

from django.http import QueryDict


class DynamicFieldsSerializerMixin:
    """
    Returns only attributes passed in `query_params['fields']`

    Eg:
        /myendpoint?fields=name,email,id
    """

    @property
    def _readable_fields(self):
        all_fields = super()._readable_fields
        query_fields = self.query_fields

        for field in all_fields:
            field_name = getattr(field, 'name', None) or getattr(field, 'field_name', None)

            if not query_fields:
                yield field
            elif query_fields and field_name in query_fields:
                yield field

    @property
    def query_fields(self):
        try:
            fields = self.context['request'].query_params['fields']
        except (KeyError, AttributeError):
            return

        return fields.split(',')


class ExcludeOnUpdateSerializerMixin:
    """
    Excludes defined fields in `Meta.exclude_on_update` when making update request
    """

    def update(self, instance, validated_data):
        # Exclude those fields defined under Meta.exclude_on_update attribute
        exclude_on_update = getattr(self.Meta, 'exclude_on_update', [])
        for field in exclude_on_update:
            validated_data.pop(field, None)

        return super().update(instance, validated_data)


class NewSerializerMixin:
    """
    A serializer mixin to create a serializer with new meta attributes.

    Eg:
    ```
    class MySerializer(NewSerializerMixin, Serializer):
        class Meta:
            fields = '__all__'

    # When it's required that we need existing serializer with different fields
    MyRefSerializer = MySerializer.New(fields=('abc', 'xyz'))
    ```
    """

    @classmethod
    def New(cls, **meta_kwargs):
        class NewCls(cls):
            class Meta(cls.Meta):
                pass

        for k, v in meta_kwargs.items():
            setattr(NewCls.Meta, k, v)

        return NewCls


class InjectReqUserSerializerMixin:
    """
    Automatically sets a model's field to request.user

    Usage:

    class MySerializer(AutoSetSerializerMixin, ModelSerializer):
        class Meta:
            ...
            model_user_field = 'user'
    """

    def to_internal_value(self, data):
        model_user_field = getattr(self.Meta, 'model_user_field')
        if model_user_field:
            if isinstance(data, QueryDict):
                data._mutable = True
            data[model_user_field] = getattr(self.context['request'].user, 'id')
            if isinstance(data, QueryDict):
                data._mutable = False
        return super().to_internal_value(data)



"""
========
Renderer
========
Change the framework's standard response structure to specific response.

The response design is follow

For single object or data::


    {
        "data": {
        }
    }

For list of object::

    {
        "meta": {
            "count": "TOTAL RECORD COUNT"
            "next": "NEXT PAGE LINK"
            "previous": "PREVIOUS PAGE LINK"
        },
        "data": [
        ]
    }

When error occurs::

    {
        "error": {
            "type": "ERROR TYPE",
            "detail": "HUMAN READABLE MESSAGE",
            "status_code": "APPROPRIATE STATUS CODE",
            "error_code": "ERROR CODE"
            "field_errors": {FIELD_NAME: [...error]},
            "non_field_errors": []
        }
    }
"""
from rest_framework.renderers import JSONRenderer as RFJSONRenderer


class JSONRenderer(RFJSONRenderer):
    """
    Override the ``render()`` of the rest framework JSONRenderer to produce JSON output as per \
    API specification
    """

    def render(self, data, accepted_media_type=None, renderer_context=None):

        response_data = {}
        context = 'data'  # Default

        # When exception handler will pass the response, we would expect
        # the context to be `error`
        if isinstance(data, dict) and data.get('_context') == 'error':
            context = 'error'
            data.pop('_context', None)

            field_errors = self._pop_field_errors(data)
            data['field_errors'] = field_errors

            # Normalizing detail field in error, only first element of dict should be visible
            data['detail'] = self._get_detail(data)
            # Always return the key
            data['non_field_errors'] = data.pop('non_field_errors', [])

        # check if the results have been paginated
        if isinstance(data, dict) and 'results' in data:
            # add the resource key and copy the results
            response_data[context] = data.pop('results')
            response_data['meta'] = data
        else:
            response_data[context] = data

        # call super to render the response
        response = super().render(response_data, accepted_media_type, renderer_context)

        return response

    def _get_detail(self, d):
        data = d.copy()

        if data.get('non_field_errors'):
            detail = data.pop('non_field_errors')[0]
        elif data.get('detail'):
            detail = data['detail']
        else:
            detail = 'Invalid request data.'

        return detail

    def _pop_field_errors(self, data):
        keys = list(data.keys())

        return {k: data.pop(k) for k in keys if k not in
                ['type', 'error_code', 'status_code', 'non_field_errors', 'detail']}



"""
===========
Middleware
===========
"""
import os
import socket
import uuid

from django.utils import timezone

branch = os.environ.get('APP_BRANCH') or 'n/a'
commit_hash = os.environ.get('APP_COMMIT_HASH') or 'n/a'


class HeadInfoMiddleware(object):
    """
    Adds these extra headers to response

    * ``X-Runtime``: Time to produce the output
    * ``X-Request-Id``: UUID to identify the request
    * ``X-Version``: Git commit hash and branch name
    * ``X-Served-By``: Host name of the machine
    """

    def __init__(self, get_response):
        self.get_response = get_response
        # One-time configuration and initialization.

    def __call__(self, request):
        # Code to be executed for each request before
        # the view (and later middleware) are called.
        start_time = timezone.now()
        request.id = uuid.uuid4().hex

        response = self.get_response(request)

        # Code to be executed for each request/response after
        # the view is called.

        response['X-Runtime'] = (timezone.now() - start_time).total_seconds()
        response['X-Request-Id'] = request.id
        response['X-Version'] = '%s#%s' % (branch, commit_hash)
        response['X-Served-By'] = socket.gethostname()
        response['Content-Length'] = len(response.content)

        return response



'''
pytz setup script
'''
import os
import re

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

me = 'Abhinav Kotak'
memail = 'in.abhi9@gmail.com'

packages = ['sbc_drf', 'sbc_drf.mixins']
package_dir = {'sbc_drf': 'src/sbc_drf'}
install_requires = open('requirements.txt', 'r').readlines()

setup(
    name='sbc_drf',
    version=re.sub(r'^v', '', os.environ.get('CI_COMMIT_TAG'), flags=re.IGNORECASE),
    zip_safe=True,
    description='A few extensions to DRF.',
    author=me,
    author_email=memail,
    maintainer=me,
    maintainer_email=memail,
    install_requires=install_requires,
    url='https://gitlab.stickboycreative.com/sbc-python/sbc_drf',
    license=open('LICENSE', 'r').read(),
    keywords=['djangorestframework', 'drf'],
    packages=packages,
    package_dir=package_dir,
    platforms=['Independant'],
    classifiers=[
        'Development Status :: 1 - beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)
